#
# install.py - SNAKE routines for setting up and performing an installation
#
# Copyright 2006, 2007 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use, modify,
# copy, or redistribute it subject to the terms and conditions of the GNU
# General Public License v.2.  This program is distributed in the hope that it
# will be useful, but WITHOUT ANY WARRANTY expressed or implied, including the
# implied warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat
# trademarks that are incorporated in the source code or documentation are not
# subject to the GNU General Public License and may only be used or replicated
# with the express permission of Red Hat, Inc. 
#

import os, gzip, tempfile
import snake.uri

# FIXME this needs to support various arches
def grubby(kernel,initrd,args='',title='snake'):
    '''Uses grubby to modify the bootloader config to make the given kernel
    and initrd the default bootloader choice. Uses 'snake' for the title if one
    is not given.'''
    # FIXME check to see if 'snake' already exists and add a suffix if needed
    cmd="""/sbin/grubby --make-default --title='%s' --add-kernel=%s --initrd=%s --args='%s'""" % (title,kernel,initrd,args)
    return os.system(cmd)
    # FIXME need to run ybin here on yaboot (Apple) machines

def __file_write(inf,outf,blocksize=4096):
    '''Write the contents of inf to outf'''
    bytes = 0
    while True:
        data = inf.read(blocksize)
        if data:
            outf.write(data)
            bytes = bytes + len(data)
        else:
            break
    inf.close()
    outf.close()
    return bytes

def add_file_to_initrd(addfile,initrd,outfile=None):
    '''Adds the given file to the given initrd. This encodes the relative
    path of the added file into the initrd, so if you want to add a file foo
    in usr/share on the initrd, it needs to be in usr/share/foo locally.'''

    # Create temporary file
    (fd,tmpinitrd) = tempfile.mkstemp('.cpio')
    os.close(fd)
    # uncompress initrd to tmpinitrd
    b = __file_write(gzip.open(initrd),open(tmpinitrd,'wb'))

    # perhaps this is a gzip'd ext2 file system?
    # NOTE pyparted works as well ... but is not included in a @base install
    tmpext2fs = tempfile.mkdtemp('.loop')
    isMounted = 0 == os.system("mount -o loop -t ext2 %s %s 2>/dev/null" % (tmpinitrd, tmpext2fs))
    if isMounted:
        add_command = "cat %(addfile)s > %(initrd)s/%(addfile)s"
    else:
        add_command = "pax -w -a -f %(initrd)s %(addfile)s"

    # add file to tmpinitrd
    r = os.system(add_command % {"initrd": isMounted and tmpext2fs or tmpinitrd, "addfile": addfile})
    if r != 0: # IT'S ALL GONE WRONG!!
        os.unlink(tmpinitrd)
        return False

    # if needed, umount ext2 initrd
    if isMounted and os.path.isdir(tmpext2fs):
        os.system("umount %s" % tmpext2fs)

    # compress tmpinitrd to outfile
    if not outfile:
        # Nothing specified - overwrite the input file
        outfile = initrd
    b = __file_write(open(tmpinitrd),gzip.open(outfile,'wb'))

    # clean up
    if os.path.isdir(tmpext2fs):
        os.rmdir(tmpext2fs)
    os.unlink(tmpinitrd)
    return True

def fetch_and_prep(uri,**kwargs):
    '''Fetches boot images from the given URI for the given arch,
    then updates the bootloader config to boot into the installer.
    Accepted kwargs:
    args=string: append given args to the boot commandline. 
    arch=string: get images for the given arch (e.g. 'xen') instead of the
                 default tree arch.
    status_callback=func(str): call the given function with various messages
                               used to indicate progress.
    ksdata=string: the given string is kickstart data, to be embedded in the
                   initrd and used during installation.'''

    if 'status_callback' in kwargs:
        status = kwargs['status_callback']
    else:
        def status(msg): pass

    if not snake.uri.check(uri):
        status("%s is not a usable SNAKE URI." % uri)
        return False

    tree = snake.uri.uri_to_tree(uri)
    if 'arch' in kwargs:
        arch = kwargs['arch']
    else:
        arch = tree.arch
    images = tree.images[arch]

    # .discinfo trees may provide a list of 'kernel' and 'initrd' images
    if not isinstance(images['kernel'], list):
        images['kernel'] = [images['kernel']]
    for kernel_img in images['kernel']:
        kernel_uri = os.path.join(uri,kernel_img)
        status("Fetching %s kernel from %s" % (arch,kernel_uri))
        try:
            kernel = snake.uri.fetchuri(kernel_uri,'/boot',False)
        except IOError, e:
            kernel = None
            continue # try the next image
        else:
            break # once we get a valid kernel_uri stop looking

    if not isinstance(images['initrd'], list):
        images['initrd'] = [images['initrd']]
    for initrd_img in images['initrd']:
        initrd_uri = os.path.join(uri,initrd_img)
        status("Fetching %s initrd from %s" % (arch,initrd_uri))
        try:
            initrd = snake.uri.fetchuri(initrd_uri,'/boot',False)
        except IOError, e:
            initrd = None
            continue # try the next image
        else:
            break # once we get a valid initrd_uri stop looking

    if not (kernel and initrd):
        status("Failed to fetch kernel or initrd")
        return False

    # If ksdata is set, but empty.. unset it.
    if 'ksdata' in kwargs and not kwargs['ksdata']:
        del kwargs['ksdata']

    if 'ksdata' in kwargs:
        status("Embedding kickstart data in initrd")
        tmpdir = tempfile.mkdtemp()
        cwd = os.getcwd()
        os.chdir(tmpdir)
        outf = open('ks.cfg','w')
        outf.write(kwargs['ksdata'])
        outf.close()
        if not add_file_to_initrd('ks.cfg',initrd):
            status("Error embedding kickstart data into initrd")
            return False

    status("Modifying bootloader config")
    # FIXME: get default args in a smarter way
    if 'args' in kwargs:
        args = kwargs['args']
    else:
        # consider also 'ipv6=auto'
        args = "ip=dhcp noipv6 lang=en keymap=us"

    if 'ksdata' in kwargs:
        args = args + " ks=file:/ks.cfg"
    else:
        args = args + " method=%s" % uri

    snake.install.grubby(kernel,initrd,args,"install %s" % str(tree))

    return True
