#
# tree.py - objects for distro trees and things like them
#
# Will Woods <wwoods@redhat.com>
#
# Copyright 2006, 2007 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use, modify,
# copy, or redistribute it subject to the terms and conditions of the GNU
# General Public License v.2.  This program is distributed in the hope that it
# will be useful, but WITHOUT ANY WARRANTY expressed or implied, including the
# implied warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat
# trademarks that are incorporated in the source code or documentation are not
# subject to the GNU General Public License and may only be used or replicated
# with the express permission of Red Hat, Inc. 
#

from snake.xmlhelper import indent, ET
import ConfigParser
import os, time

class ComposeTarget(object):
    '''A generic base class for trees and other things built in a compose'''
    def __init__(self,**kwargs):
        self.id=""       # a unique ID string that identifies this object
        self.name=""     # a human-readable string to identify this tree
        self.compose=""  # compose ID
        self.time=0.0    # timestamp
        self.family=""   # family (e.g. Fedora, Red Hat Enterprise Linux)
        self.variant=""  # variant/flavor (e.g. Desktop, AS)
        self.version=""  # tree version. FIXME: what about "4.5.1"?
        self.arch=""     # arch
        self.reldir=""   # (optional) location relative to parent compose
        self.uris=[]  # list of URIs for this particular tree
        # convenience - what type am I?
        #self.type=type(self).__name__
        if 'xml' in kwargs:
            self.parsexml(kwargs['xml'])
        elif 'dir' in kwargs:
            self.loaddir(kwargs['dir'])
        elif 'dict' in kwargs:
            self.__dict__ = kwargs['dict']

    def loaddir(self,dir):
        xmlfile=os.path.join(dir,'.%s.xml' % type(self).__name__.lower())
        treeinfo=os.path.join(dir,'.treeinfo')
        discinfo=os.path.join(dir,'.discinfo')
        if os.path.exists(xmlfile):
            return self.parsexml(xmlfile)
        elif os.path.exists(treeinfo):
            return self.parsetreeinfo(treeinfo)
        elif os.path.exists(discinfo):
            return self.parsediscinfo(discinfo)

    def parsetreeinfo(self,treeinfo):
        '''Parse a .treeinfo file to fill in object attributes.
           Argument may be either the path to a file, or an file object.'''
        c=ConfigParser.ConfigParser()
        if hasattr(treeinfo,'readline'):
            c.readfp(treeinfo) # read file-like object
        else:
            c.read(treeinfo)   # try to read it as if it was a filename
        self.time = float(c.get('general','timestamp'))
        self.arch = c.get('general','arch')
        self.family  = c.get('general','family')
        self.version = c.get('general','version')
        self.variant = c.get('general','variant')
        if c.has_option('general','name'):
            self.name = c.get('general','name')
        if c.has_option('general','id'):
            self.id = c.get('general','id')
        if c.has_option('general','uris'):
            self.uris = c.get('general','uris').split()
        self._fill_in_blanks()
        self.config = c
        return c

    def _fill_in_blanks(self):
        '''Fill in missing 'name' and 'id' attributes by using the timestamp
           and _makename methods'''
        if not self.id:
            self.id = str(self.time)
        if not self.name:
            self.name = self._makename()

    def parsediscinfo(self,discinfo):
        '''Parse a .discinfo file to fill in some object attributes.
           (This kind of sucks since it lacks images/stage2 info - use
           parsetreeinfo instead if possible).
           Argument may be a path to a file, or an open file object.'''
        if hasattr(discinfo,'readline'):
            fp = discinfo
        else:
            fp = open(discinfo)
        self.time = float(fp.readline())
        # Figure out family and version
        desc = fp.readline() # e.g. "Fedora Core 6"
        if desc.startswith("Fedora"):
            self.family = "Fedora"
        elif desc.startswith("Red Hat Enterprise Linux"):
            self.family = "Red Hat Enterprise Linux"
            var = desc.lstrip(self.family).split()[0]
            if not var.isdigit():
                self.variant = var
        ver = desc.split()[-1]
        if ver.isdigit():
            self.version = ver
        # Figure out the arch
        self.arch = fp.readline().strip()
        # Fill in blanks (.name/.id)
        self._fill_in_blanks()
        return fp

    def parsexml(self,xml):
        elemtree=ET.parse(xml)
        r=elemtree.getroot()
        self.compose=r.find('compose').text
        self.time=float(r.find('time').text)
        self.arch=r.find('arch').text
        self.family=r.find('family').text
        self.version=r.find('version').text
        if r.find('variant'):
            self.variant = r.find('variant').text

        # FIXME: we need to settle on having the name as an xml attribute or element ... this attempts to handle both
        self.name = r.get('name')
        if not self.name and r.find('name'):
            self.name = r.find('name').text
        if not self.name: 
            self.name = self._makename()
        return elemtree

    def element(self,klist=()):
        root=ET.Element(type(self).__name__.lower(),name=self.name)
        for k,v in self.__dict__.items():
            if (k in klist+('compose','time','arch')) and v:
                e=ET.SubElement(root,k)
                e.text=str(v)
        return root

    def xml(self):
        return ET.tostring(indent(self.element()))

    def _makename(self):
        if self.compose and hasattr(self.compose,'name'):
            list = [self.compose.name, self.arch]
        else:
            list = [self.family]
            if self.variant and self.variant != self.family:
                list.append(self.variant)
            list.append(self.version)
        return '/'.join(list)

    def __str__(self):
        timestamp = time.localtime(float(self.time))
        time_str = time.strftime("%Y%m%d",timestamp)
        r = "%s %s " % (self.compose or self.family, self.version)
        if self.variant and self.variant != self.family: r += self.variant + " "
        r = r + "%s (%s)" % (self.arch, time_str)
        return r

class Tree(ComposeTarget):
    '''A class that represents an installable tree'''
    # This subclass adds the 'images' dict, which contains boot images
    def __init__(self,**kwargs):
        self.images={}
        ComposeTarget.__init__(self,**kwargs)

    def parsetreeinfo(self,treeinfo):
        c=ComposeTarget.parsetreeinfo(self,treeinfo)
        for s in c.sections():
            if s.startswith('images-'):
                self.images[s.replace('images-','')]=dict(c.items(s))

    def parsediscinfo(self,discinfo):
        ComposeTarget.parsediscinfo(self,discinfo)
        # Fill in the 'images' dicts with the standard paths. Hope they're
        # right. This is why discinfo kinda sucks.
        x86_images = {'initrd':'images/pxeboot/initrd.img',
                      'kernel':'images/pxeboot/vmlinuz',
                      'diskboot.img':'images/diskboot.img',
                      'boot.iso':'images/boot.iso'}
        xen_images = {'initrd':'images/xen/initrd.img',
                      'kernel':'images/xen/vmlinuz'}
        ia64_images = {'initrd':'images/pxeboot/initrd.img',
                       'kernel':'images/pxeboot/vmlinuz',
                       'boot.img':'images/boot.img',
                       'boot.iso':'images/boot.iso'}
        ppc_images = {'ppc':  {'initrd':'ppc/ppc32/ramdisk.image.gz',
                               'kernel':'ppc/ppc32/vmlinuz',
                               'zimage':'images/netboot/ppc32.img'},
                      'ppc64':{'initrd':['ppc/ppc64/ramdisk.image.gz', 
                                         'ppc/chrp/ramdisk.image.gz'],
                               'kernel':['ppc/ppc64/vmlinuz', 
                                         'ppc/chrp/vmlinuz', 
                                         'ppc/chrp/vmlinux'],
                               'zimage':['images/netboot/ppc64.img', 
                                         'images/pseries/netboot.img', 
                                         'images/netboot.img']}}
        s390_images = {'initrd':'images/initrd.img',
                       'kernel':'images/kernel.img',
                       'generic.prm':'images/generic.prm',}

        self.images = dict()
        # We can't actually check for the existence of these files.
        # You better handle missing file errors yourself, buddy.
        if self.arch in ('i386', 'x86_64'):
            self.images = {self.arch:x86_images,'xen':xen_images}
        elif self.arch == 'ia64':
            self.images = {self.arch:ia64_images,'xen':xen_images}
        elif self.arch == 'ppc':
            self.images = ppc_images
        elif self.arch.startswith("s390"):
            self.images = {self.arch:s390_images}

    def parsexml(self,xml):
        elemtree=ComposeTarget.parsexml(self,xml)
        r=elemtree.getroot()
        # FIXME: This doesn't work!!
        for image_elem in r.findall('images'):
            arch=image_elem.get('arch')
            self.images[arch]=dict()
            for e in image_elem.findall('file'):
                self.images[arch][e.get('type')] = e.text

    def element(self):
        root=ComposeTarget.element(self,('family','version','variant'))
        for arch,files in self.images.items():
            image_elem=ET.SubElement(root,'images',arch=arch)
            for type,path in files.items():
                e=ET.SubElement(image_elem,'file',type=type)
                e.text=path
        return root

class IsoSet(Tree):
    '''A class that represents a set of ISO images'''
    # This class has a list of iso images, and the boot images live therein.
    # FIXME: We'll need an ISO helper to pull info out of the disc images.
    pass

# Source and Debug are two other things we might need to consider here.
# These are just stubs, defined so we don't forget about them.

class Source(ComposeTarget):
    '''A class for a SRPM dir'''
    pass

class Debug(ComposeTarget):
    '''A class for a debuginfo dir'''
    pass
