#
# uri.py - misc. routines for handing URIs to trees etc.
#
# Copyright 2006, 2007 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use, modify,
# copy, or redistribute it subject to the terms and conditions of the GNU
# General Public License v.2.  This program is distributed in the hope that it
# will be useful, but WITHOUT ANY WARRANTY expressed or implied, including the
# implied warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat
# trademarks that are incorporated in the source code or documentation are not
# subject to the GNU General Public License and may only be used or replicated
# with the express permission of Red Hat, Inc. 
#

import sys, os, urllib2, urlparse, tempfile
from snake.tree import Tree

# URI errors are basically a failure to communicate with a remote machine, so
# this is subclassed from IOError. urllib2.URLError is also an IOError subclass.
class URIError(IOError): pass

# We cache the contents of /etc/mtab ... the following variables are used 
# to keep our cache in sync
mtab_mtime = None
mtab_nfsmap = []

# Define a helper object to be used by older python versions
class UriDummyType: 
    def __init__(self, args):
        (self.scheme, self.hostname, self.path, self.query, self.fragment) = args
    def geturl(self):
        return urlparse.urlunsplit((self.scheme, self.hostname, self.path, self.query, self.fragment))

def uriopen(uri):
    '''Open a URL and return a file-like object.'''
    fd = None
    u = parse(uri)
    if u.scheme in ('http','https','ftp'):
        try:
            fd = urllib2.urlopen(uri)
        except IOError, e:
            raise URIError, "urllib error: %s" % e
        if uri.startswith('ftp') and not fd.info():
            raise URIError, "urllib error: file not found"
    elif u.scheme == 'nfs':
        # is this server already mounted?
        (netloc,path) = u.path.split(':',1)
        for mount in find_nfs_mounts():
            if mount['hostname'] == netloc and \
                path.startswith(mount['path']):
                local_uri = "file://%s" % path.replace(mount['path'],mount['file'])
                try:
                    fd = urllib2.urlopen(os.path.join(local_uri))
                except OSError, e:
                    raise URIError, "urllib error: %s" % e
                break
        # try to nfs mount manually
        if not fd:
            '''FIXME - call os.system('mount -t nfs ...') directly'''
            raise URIError, "NFS mount not implemented"
    else:
        raise URIError, "Unimplemented scheme %s" % u.scheme
    return fd

def fetchuri(uri,target,overwrite=True):
    '''Fetches the file given by the URI into the target dir.
    If overwrite is false, it will append ".1" (or similar) to the filename
    if it already exists.'''
    targetfile = os.path.join(target,os.path.basename(uri))
    if not overwrite and os.path.exists(targetfile):
        count = 1
        newfile = "%s.%i" % (targetfile,count)
        while os.path.exists(newfile):
            count = count + 1
            newfile = "%s.%i" % (targetfile,count)
        targetfile = newfile
    data = uriopen(uri)
    f = open(targetfile,"w")
    f.write(data.read())
    data.close()
    f.close()
    return targetfile

def parse(uri):
    '''Returns true if uri is a usable SNAKE uri'''
    uri_obj = urlparse.urlsplit(uri)

    if sys.version_info < (2,5):
        '''Older python returns a tuple, let's fake a class'''
        uri_obj = UriDummyType(uri_obj)

    if (uri_obj.scheme in ('http','https','ftp') and \
        not (uri_obj.path or uri_obj.hostname)) or \
       (uri_obj.scheme in ('nfs') and 
        not uri_obj.path):
        raise URIError, 'Unusable URI: %s' % uri
    return uri_obj

def find_nfs_mounts(mtab="/etc/mtab"):
    '''Scan /etc/mtab and return list of nfs mounts'''
    global mtab_mtime, mtab_nfsmap

    mtab_stat = os.stat(mtab)
    if mtab_stat.st_mtime != mtab_mtime:
        mtab_mtime = mtab_stat.st_mtime
        f = open(mtab)
        mtab_nfsmap = []
        for line in f:
            (spec,file,vfstype,mntops,x,y) = line.split()
            if vfstype == 'nfs': 
                (hostname,path) = spec.split(':')
                mtab_nfsmap.append({"spec": spec, "hostname": hostname, "path": path, 
                               "file": file, "vfstype": vfstype, "mntops": mntops })
        f.close()
    return mtab_nfsmap

def nfs_to_local(uri_obj):
    '''convert an nfs: uri into a local file:// uri'''
    local_uri = ""
    netloc = uri_obj.path.split(':',1)[0]
    # is the server already mounted?
    for mount in find_nfs_mounts():
        if mount['hostname'] == netloc and \
            uri_obj.path.startswith(mount['path']):
            local_uri = "file://%s" % uri_obj.path.replace(mount['path'],mount['file'])
            break
    # try to nfs mount manually
    if not local_uri: 
        '''attempt to nfs mount old-school'''
        raise IOError, "NFS mount not supported"
        # tmpdir = tempfile.mkdtemp()
        # cmd = '''mount -t nfs %s:%s %s''' % (uri_obj.hostname, uri_obj.path, tmpdir)
        # rc = os.system(cmd)
        # local_uri = "file://%s" % tmpdir
    return local_uri

def treeinfo(uri):
    '''Returns a file-like object for the .treeinfo in the given URI'''
    return uriopen(os.path.join(uri,".treeinfo"))

def discinfo(uri):
    '''Returns a file-like object for the .discinfo in the given URI'''
    return uriopen(os.path.join(uri,".discinfo"))

def uri_to_tree(uri):
    '''Return a Tree object representing the tree at the given URI'''
    t = Tree()
    try:
        t.parsetreeinfo(treeinfo(uri))
    except URIError:
        t.parsediscinfo(discinfo(uri))
    return t

def check(uri):
    '''Check the validity of a given URI'''
    try:
        t = treeinfo(uri)
    except URIError:
        try:
            t = discinfo(uri)
        except URIError:
            return False
    t.close()
    return True
